import asyncio
import time
import os

from hedra.reporting import Handler
from hedra.runners.leader_services.bootstrap import BootstrapManager
from hedra.runners.leader_services.leader_registry import LeaderRegistry
from hedra.runners.leader_services.leader_registry.leader_service import LeaderService

class ElectionManager:


    def __init__(self) -> None:

        logger = Logger()
        self.session_logger = logger.generate_logger()
        self.expected_leaders = []
        self.polls_received = 0
        self.is_leader = False
        self.leader_ip = None
        self.leader_port = None
        self.leader_address = None
        self.leader_id = None
        self.registered_services = None
        self.leader_election_timeout = None
        self.discovery_mode = None
        self.loop = asyncio.get_event_loop()
        
    async def setup(self, config, leader_id):
        self.config = config
        self.reporter_config = config.reporter_config

        self.leader_id = leader_id

        self.leader_election_timeout = config.distributed_config.get(
            'leader_election_timeout', 
            60
        )

        self.elected_leader_id = self.leader_id
        self.leaders_count = len(self.expected_leaders)
        self.elected_leader = None
        self.poll_completed = False

    async def register(self, leader_registration_request):
        await self.registered_services.register(leader_registration_request)
        return {
            'host_address': self.leader_address,
            'host_id': str(self.leader_id)
        }

    async def elect_leader_and_submit(self, job_results, job_id=None):

        handler = Handler(self.config)
        await handler.initialize_reporter()

        if job_id:
            self.session_logger.info(f'\nJob - {job_id} has completed...')

        else:
            self.session_logger.info('\nAll workers have signaled completion, ending session...')

        submitted = False
        elapsed = 0

        start = time.time()
        overall_timeout = self.config.distributed_config.get('leader_results_timeout', 60)
        
        if self.leaders_count > 0:
            while submitted is False and elapsed < overall_timeout:

                registration_complete = await self.registered_services.registration_complete()
                if registration_complete is False:
                    self.session_logger.info('\nElecting leader to submit results...')
                    await self.register_leaders()
                    await self.wait_until_done()

                self.elected_leader = await self.elect_leader()

                if self.elected_leader:
                    self.session_logger.info(f'Elected leader at - {self.elected_leader.service_address} - to submit results.')

                if self.is_leader:
                    await handler.merge(job_results)
                    stats = await handler.get_stats()

                    self.session_logger.info(f'Leader - {self.leader_address} - submitting session results...')
                    submitted = await handler.submit(stats)
                    self.session_logger.info(f'Results submitted!\n')
                    self.poll_completed = submitted
                    await self.wait_until_poll_complete()

                else:

                    completed = await self.poll_leader()
                    if completed:
                        submitted = True

                        self.session_logger.info('Poll complete.\n')

                    else:
                        await self.remove_elected_leader()

                        self.session_logger.info(f'Leader at - {self.elected_leader.service_address} - failed to respond. Electing new leader.')

                elapsed = time.time() - start

        if submitted is False:
            self.session_logger.info(f'Elected leader at - {self.leader_address} - to submit results.\n')
            await handler.merge(job_results)
            stats = await handler.get_stats()
            await handler.submit(stats)
            self.session_logger.info(f'Results submitted!\n')

    async def elect_leader(self):

        registration_complete = await self.registered_services.registration_complete()

        if registration_complete:
            elected_leader = await self.registered_services.find_min_id()
            await self.registered_services.set_elected_leader(elected_leader)
            self.elected_leader_id = self.registered_services.elected_leader.service_id

            if self.elected_leader_id.int == self.leader_id.int:
                self.is_leader = True

        return self.registered_services.elected_leader

    async def remove_elected_leader(self):
        if self.registered_services.elected_leader:
            await self.registered_services.remove_service(self.registered_services.elected_leader.service_address)

    async def wait_until_done(self):
        elapsed = 0
        start = time.time()
        wait_interval = self.config.distributed_config.get('leader_wait_interval', 1)

        registration_complete = await self.registered_services.registration_complete()

        while registration_complete is False and elapsed < self.leader_election_timeout:
            registration_complete = await self.registered_services.registration_complete()
            await asyncio.sleep(wait_interval)
            elapsed = time.time() - start

        return await self.registered_services.registration_complete()

    async def register_leaders(self, config):
        await self._discover_leaders(config)
        
        async for service in self.registered_services.services():
            is_registered = await service.is_registered(self.registered_services)
            
            if is_registered is False:
                response = await service.register(self.leader_address, self.leader_id)
                if response:
                    await self.registered_services.register_by_response(response)

    async def setup_leader_addresses(self, config):
        leader_ips = config.distributed_config.get(
            'leader_ips',
            os.getenv('LEADER_IPS')
        )

        leader_ports = config.distributed_config.get(
            'leader_ports',
            os.getenv('LEADER_PORTS', '6669')
        )

        leader_addresses = config.distributed_config.get(
            'leader_addresses',
            os.getenv('LEADER_ADDRESSES')
        )

        self.discovery_mode = config.distributed_config.get(
            'discovery_mode',
            os.getenv('DISCOVERY_MODE', 'static')
        )

        if leader_addresses:
            self.expected_leaders = leader_addresses.split(',')
            leader_ips = []
            leader_ports = []

            for leader_address in self.expected_leaders:
                leader_ip, leader_port = leader_address.split(':')
                leader_ips.append(leader_ip)
                leader_ports.append(int(leader_port))

        elif leader_ips:
            leader_ips = leader_ips.split(',')
            leader_ports = leader_ports.split(',')

            for ip, port in zip(leader_ips, leader_ports):
                self.expected_leaders.append(f'{ip}:{port}')
        
        else:

            if leader_ips is None:
                leader_ips = ['localhost']

            leader_ports = leader_ports.split(',')
            self.expected_leaders = [f'{leader_ips[0]}:{leader_ports[0]}']

        self.leader_ip = leader_ips.pop()
        self.leader_port = leader_ports.pop()
        self.leader_address = self.expected_leaders.pop()

    async def _discover_leaders(self, config):

        bootstrap_manager = BootstrapManager(config)
        self.registered_services = LeaderRegistry()

        if self.discovery_mode == 'kubernetes':
            self.session_logger.info('\nBeginning dynamic Kubernetes leader discovery.')

            await bootstrap_manager.client.discover(
                host_address=self.leader_address
            )

            self.session_logger.info('Leader discovery complete!\n')

            discovered = await bootstrap_manager.discovered()
            await self.registered_services.setup(discovered)
            self.leader_id = self.registered_services.leader_id

        elif self.discovery_mode == 'dynamic':
            self.session_logger.info('\nBeginning dynamic leader discovery.')

            await bootstrap_manager.register_leader(self.leader_address)
            await bootstrap_manager.wait_for_leaders(
                self.leader_address,
                self.leader_id
            )

        else:
            host = LeaderService(
                self.leader_address,
                self.leader_id,
                host=True
            )

            services = [
                LeaderService(
                    service_address,
                    None,
                    host=False
                ) for service_address in self.expected_leaders
            ]

            await self.registered_services.setup([
                *services,
                host
            ])
    
    async def wait_until_poll_complete(self):
        poll_timeout = self.config.distributed_config.get('leader_poll_timeout', 15)
        wait_interval = self.config.distributed_config.get('leader_wait_interval', 1)

        elapsed = 0
        start = time.time()

        while self.polls_received < self.leaders_count and elapsed < poll_timeout:
            await asyncio.sleep(wait_interval)

            elapsed = time.time() - start
                    
    async def poll_leader(self):
        poll_timeout = self.config.distributed_config.get('leader_poll_timeout', 15)
        wait_interval = self.config.distributed_config.get('leader_wait_interval', 1)

        elapsed = 0
        start = time.time()
        completed = False

        while completed is False and elapsed < poll_timeout:
            if self.registered_services.registered_count <= 1:
                return False

            response = await self.elected_leader.poll()
            if response:
                completed = response.completed

            await asyncio.sleep(wait_interval)
            elapsed = time.time() - start

        return completed





from .context_group import ContextGroup
from .context_config import ContextConfig


class ContextPool:

    __slots__ = (
        'size',
        'group_size',
        'group_count',
        'contexts'
    )

    def __init__(self, pool_size, group_size) -> None:
        self.size = pool_size
        self.group_size = group_size
        self.groups_count = int(pool_size/group_size)
        self.contexts = []

    def __iter__(self):
        for context_group in self.contexts:
            yield context_group

    async def __aiter__(self):
        for context_group in self.contexts:
            yield context_group

    def create_pool(self, config: ContextConfig):
        self.contexts = [
            ContextGroup(
                **config.data,
                concurrency=self.group_size
            ) for _ in range(self.groups_count)
        ]


import grpc
import os
import psutil
from .service import LeaderService
from .proto import add_DistributedServerServicer_to_server


class LeaderServer:

    def __init__(self, config):
        self.config = config
        self.ip = None
        self.port = None
  
        self.session_logger.info('Initializing leader server...')
        self._worker_threads = config.distributed_config.get(
            'max_workers', 
            os.getenv(
                'MAX_THREAD_WORKERS',
                psutil.cpu_count(logical=False)
            )
        )

        self.session_logger.debug('Generating server with - {threads} - threads'.format(threads=self._worker_threads))
        self.server = grpc.aio.server()

        self.service = LeaderService()

        add_DistributedServerServicer_to_server(
            self.service,
            self.server
        )


        self.run_server = False

    async def create_server(self, config):
        await self.service.setup_leader_addresses(config)
        self.ip = self.service.leader_ip
        self.port = self.service.leader_port

        server_address = '[::]:{port}'.format(port=self.port)
        self.server.add_insecure_port(server_address)
        await self.server.start()

    async def setup(self, leader_id):
        await self.service.setup(
            self.config,
            leader_id
        )

    async def run(self):
        self.session_logger.info('Leader server serving on port - {port}'.format(port=self.port))
        
        workers = self.service.job_registry.workers.registry.worker_addresses

        self.session_logger.info(f'Initializing with: - {workers} - workers.')
        if len(self.service.job_registry.workers.registry.worker_addresses) > 0:
            await self.service.job_registry.workers.initialize()

        self.service.running = True
        self.run_server = True

    async def stop_server(self):
        await self.server.stop(0)

    async def add_new_workers(self, new_workers_queue):
        await self.service.job_registry.workers.add_new_workers(new_workers_queue)

    async def leader_register_workers(self):
        if len(self.service.job_registry.workers.registry.worker_addresses) > 0:
            await self.service.job_registry.workers.leader_register_worker()

    async def start_monitor(self):
        await self.service.job_registry.start_job_monitor()

    async def stop_monitor(self):
        await self.service.job_registry.stop_job_monitor()




from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    DogStatsDConfig
)


class SubmitDogStatsDResultsStage(Submit):
    config=DogStatsDConfig(
        host='localhost',
        port=8125
    )


import asyncio
import dill
import traceback

from pycli_tools.arguments.bundler import Bundler
from hedra.core import Executor
from hedra.core.graphs.hooks import (
    Execute
)
from hedra.core.graphs.stages.base.stage import Stage


def run_job(config):
    logger = Logger()
    session_logger = logger.generate_logger()

    try:
        
        import asyncio
        
        policy = asyncio.get_event_loop_policy()
        policy.set_event_loop(policy.new_event_loop())
        loop = asyncio.get_event_loop()

        config = dill.loads(config)

        job_config = config.get('config')

        if job_config.executor_config.get('engine_type') == 'action-set':
            bundler = Bundler(options={
                'class_type': Stage,
                'package_name': job_config.executor_config.get('import_filepath')
            })


            discovered = bundler.discover()
      
            actions = {}

            for action_set in discovered:
                actions[action_set.__name__] = action_set

            job_config.actions = actions.values()

        worker = Executor(job_config)
        results = loop.run_until_complete(_run_job(worker))
        
        return {
            'stats': worker.pipeline.stats,
            'events': results
        }

    except Exception as err:
        session_logger.info(f'{traceback.format_exc()}')
        session_logger.info(f'Parallel pipeline encountered exception - {str(err)}')

async def _run_job(worker):
    await worker.setup()

    for pipeline_stage in worker.pipeline:
        stage.value = pipeline_stage.name
        await worker.pipeline.execute_stage(pipeline_stage)

    stage.value = b'serializing'
    parsed_results = await worker.calculate_results()

    return parsed_results

def create_job(sync_barrier, process_queue, process_stage, optimal_aps, optimal_batch_size, optimal_batch_time):
    global process_barrier
    process_barrier = sync_barrier

    global processed
    processed = process_queue

    global stage
    stage = process_stage

    global optimized_aps
    optimized_aps = optimal_aps

    global optimized_batch_size
    optimized_batch_size = optimal_batch_size

    global optimized_batch_time
    optimized_batch_time = optimal_batch_time



import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    TimescaleDBConfig
)


class SubmitTimescaleDBResultsStage(Submit):
    config=TimescaleDBConfig(
        host='127.0.0,1',
        database='results',
        username=os.getenv('TIMESCALEDB_USERNAME', ''),
        password=os.getenv('TIMESCALEDB_PASSWORD', ''),
        events_table='events',
        metrics_table='metrics'
    )


import asyncio
import functools
import re
import psutil
import uuid
from typing import List
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet


try:
    import newrelic.agent
    from .newrelic_config import NewRelicConfig
    has_connector=True

except Exception:
    newrelic = None
    NewRelicConfig = None
    has_connector=False


class NewRelic:

    def __init__(self, config: NewRelicConfig) -> None:
        self.config_path = config.config_path
        self.environment = config.environment
        self.registration_timeout = config.registration_timeout
        self.shutdown_timeout = config.shutdown_timeout or 60
        self.newrelic_application_name = config.newrelic_application_name

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self.client = None
        self._loop = asyncio.get_event_loop()

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to NewRelic - Using config at path - {self.config_path} - Environment - {self.environment}')

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                newrelic.agent.initialize,
                config_file=self.config_path,
                environment=self.environment
            )
        )

        self.client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                newrelic.agent.register_application,
                name=self.newrelic_application_name,
                timeout=self.registration_timeout
            )
        )

        await asyncio.sleep(1)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to NewRelic - Using config at path - {self.config_path} - Environment - {self.environment}')

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to NewRelic')

        for event in events:
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.record_custom_event,
                    event.name,
                    event.record
                )
            )
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to NewRelic')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to NewRelic')
        
        for metrics_set in metrics_sets:  
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for field, value in metrics_set.common_stats.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metric - {metrics_set.name}:common:{field}')
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.record_custom_metric,
                        f'{metrics_set.name}_{field}',
                        value
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to NewRelic')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to NewRelic')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                for field, value in group.stats.items():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metric - {metrics_set.name}:{group_name}:{field}')
                    await self._loop.run_in_executor(
                        self._executor,
                        functools.partial(
                            self.client.record_custom_metric,
                            f'{metrics_set.name}_{group_name}_{field}',
                            value
                        )
                    )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to NewRelic')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to NewRelic')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_group_name, group in metrics_set.custom_metrics.items():
                for field, value in group.items():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metric - {metrics_set.name}:{custom_group_name}:{field}')
                    await self._loop.run_in_executor(
                        self._executor,
                        functools.partial(
                            self.client.record_custom_metric,
                            f'{metrics_set.name}_{custom_group_name}_{field}',
                            value
                        )
                    )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to NewRelic')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to NewRelic')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                error_message = re.sub(
                    '[^0-9a-zA-Z]+', 
                    '_',
                    error.get(
                        'message'
                    ).lower()
                )

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.record_custom_metric,
                        f'{metrics_set.name}_errors_{error_message}',
                        error.get('count')
                    )
                )  
                
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to NewRelic')          

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connectiion to NewRelic')

        await self._loop.run_in_executor(
            self._executor,
            self.client.shutdown
        )

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Session Closed - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connectiion to NewRelic')


from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class HoneycombConfig(BaseModel):
    api_key: str
    dataset: str
    reporter_type: ReporterTypes=ReporterTypes.Honeycomb


from typing import List


class InvalidActionError(Exception):

    def __init__(self, specified_workflow_action: str, valid_workflow_actions: List[str]) -> None:

        valid_action_names = '\n-'.join(valid_workflow_actions)

        super().__init__(
            f'\n\nError - Invalid workflow action - {specified_workflow_action} - specified. Valid actions are:\n\n{valid_action_names}\n'
        )


import asyncio
from weakref import ref
from hedra.core.engines.types.common.protocols.shared.reader import Reader
from .protocol import TCPProtocol


class TLSProtocol(TCPProtocol):

    def upgrade_reader(self, reader: Reader):

        if self._stream_reader:
            self._stream_reader.set_exception(Exception('upgraded connection to TLS, this reader is obsolete now.'))

        self._stream_reader_wr = ref(reader)
        self._source_traceback = reader._source_traceback


import asyncio
import traceback
from typing import Dict, List, Union, Iterator
from types import FunctionType
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.http2 import(
    MercuryHTTP2Client,
    HTTP2Action,
    HTTP2Result
)
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common import Timeouts
from hedra.core.engines.client.store import ActionsStore
from hedra.logging import HedraLogger
from .base_client import BaseClient


class HTTP2Client(BaseClient[MercuryHTTP2Client, HTTP2Action, HTTP2Result]):
    
    def __init__(self, config: Config) -> None:
        super().__init__()

        self.session = MercuryHTTP2Client(
            concurrency=config.batch_size,
            timeouts=Timeouts(
                connect_timeout=config.connect_timeout,
                total_timeout=config.request_timeout
            ),
            reset_connections=config.reset_connections
        )
        self.request_type = RequestTypes.HTTP2
        self.client_type = self.request_type.capitalize()

        self.actions: ActionsStore = None
        self.next_name = None
        self.intercept = False

        self.logger = HedraLogger()
        self.logger.initialize()

    def __getitem__(self, key: str):
        return self.session.registered.get(key)

    async def get(
        self, 
        url: str, 
        headers: Dict[str, str] = {}, 
        user: str = None,
        tags: List[Dict[str, str]] = []
    ):

        request = HTTP2Action(
            self.next_name,
            url,
            method='GET',
            headers=headers,
            data=None,
            user=user,
            tags=tags
        )
        
        return await self._execute_action(request)

    async def post(
        self,
        url: str, 
        headers: Dict[str, str] = {}, 
        data: Union[dict, str, bytes, Iterator] = None,
        user: str = None,
        tags: List[Dict[str, str]] = []
    ):

        request = HTTP2Action(
            self.next_name,
            url,
            method='POST',
            headers=headers,
            data=data,
            user=user,
            tags=tags
        )
    
        return await self._execute_action(request)


    async def put(
        self,
        url: str, 
        headers: Dict[str, str] = {}, 
        data: Union[dict, str, bytes, Iterator] = None,
        user: str = None,
        tags: List[Dict[str, str]] = []
    ):

        request = HTTP2Action(
            self.next_name,
            url,
            method='PUT',
            headers=headers,
            data=data,
            user=user,
            tags=tags
        )

        return await self._execute_action(request)


    async def patch(
        self,
        url: str, 
        headers: Dict[str, str] = {}, 
        data: Union[dict, str, bytes, Iterator] = None,
        user: str = None,
        tags: List[Dict[str, str]] = []
    ):

        request = HTTP2Action(
            self.next_name,
            url,
            method='PATCH',
            headers=headers,
            data=data,
            user=user,
            tags=tags
        )

        return await self._execute_action(request)


    async def delete(
        self, 
        url: str, 
        headers: Dict[str, str] = {}, 
        user: str = None,
        tags: List[Dict[str, str]] = []
    ):

        request = HTTP2Action(
            self.next_name,
            url,
            method='DELETE',
            headers=headers,
            data=None,
            user=user,
            tags=tags
        )

        return await self._execute_action(request)



from .worker_pb2 import (
    LeaderUpdateRequest,
    WorkerServerUpdateResponse,
    NewJobRequest,
    NewJobResponse,
    PollWorkerRequest,
    PollWorkerResponse,
    JobCompleteRequest,
    JobCompleteResponse,
    WorkerHeartbeatRequest,
    WorkerHeartbeatResponse,
    LeaderRegisterWorkerRequest,
    LeaderRegisterWorkerResponse
)


from .worker_pb2_grpc import (
    WorkerServerServicer,
    WorkerServerStub,
    add_WorkerServerServicer_to_server
)


import time


class JobPipeline:

    def __init__(self, job_id, leader_address=None, leader_id=None, timeout=600) -> None:
        self.job_id = job_id
        self.reported_completions =  {
            'setup': 0,
            'warmup': 0,
            'optimize': 0,
            'execute': 0,
            'results': 0,
            'done': 0
        }

        self.stages_count = len(self.reported_completions)
        self.workers_count = 0
        self.current_stage = 'setup'
        self.current_stage_completed = False
        self.completed = False
        self.start = None
        self.elapsed = 0
        self.timeout = timeout
        self.workers = []
        self.excluded_workers = []
        self.leader_address = leader_address
        self.leader_id = leader_id

        self.config = None
        self.configs = {}

    async def setup_pipeline(self, new_job, workers, group_on='server'):
        self.config = new_job
        self.workers_count = len(workers)

        if self.workers_count > 0:

            batch_size = self.config.get('batch_size', 1000)

            job_size = int(batch_size / self.workers_count)
            remainder = 0
            job_sizes = {}

            if batch_size % self.workers_count:
                remainder = batch_size % self.workers_count

            for idx, worker in enumerate(workers):
    
                job_sizes[worker] = job_size

                if idx == (self.workers_count - 1):
                    job_sizes[worker] += remainder

            for worker in workers:
                config = dict(self.config)
                config['batch_size'] = job_sizes[worker]
                self.configs[worker] = config

            self.expected_workers = workers

    async def get_job_config(self, worker):
        if worker.worker_addresses in self.worker_configs:
            return self.worker_configs[worker.worker_address]

        return None

    async def start_timer(self):
        self.start = time.time()

    async def update_job_runtime(self):
        self.elapsed = time.time() - self.start

    async def check_if_timeout_exceeded(self):
        return self.elapsed > self.job_timeout

    async def update_completed(self, job_stage_completed):
        self.reported_completions[job_stage_completed.stage_completed] += 1
        stage_completions = self.reported_completions.get(job_stage_completed.stage_completed, 0)
        if stage_completions >= self.workers_count:
            self.current_stage_completed = True

        else:
            self.current_stage_completed = False

    async def check_if_completed(self):
        self.completed = self.reported_completions['done'] >= len(self.workers) and self.reported_completions['done'] > 0
        return self.completed

    async def set_active_workers(self, active_count):
        self.workers_count = active_count

    async def update_pipeline_workers(self, current_workers, failed_workers):
        self.excluded_workers += failed_workers
        self.excluded_workers = list(set(self.excluded_workers))
        self.workers = [worker for worker in current_workers if worker not in self.excluded_workers]

    async def update_pipeline_status(self, updated_pipeline_status):
        new_pipeline = {}

        for stage in updated_pipeline_status:
            current_reported_completions = self.reported_completions[stage]
            updated_reported_completions = int(updated_pipeline_status[stage])

            new_pipeline[stage] = max(current_reported_completions, updated_reported_completions)

        self.reported_completions = new_pipeline

        return new_pipeline





import functools
from .hook_type import HookType
from hedra.core.graphs.hooks.registry.registrar import registrar


@registrar(HookType.SETUP)
def setup(metadata={}):
    '''
    Setup Hook

    The @setup(<name>, ...) decorator is an optional means of specifying that
    a method of a class inherting ActionSet should executor prior to testing as a means
    of setting up/initializing test state. You may specify as many setup hooks
    as needed, however all methods wrapped in the decorator must be asynchronous
    and return valid results.

    You may pass the following to a setup hook:

    - name (required:string - positional)
    - group (optional:string - keyword)
    - metadata (optional:dict - keyword - may include env, user, type, tags [list], and url)
    
    '''
    def wrapper(func):
 
        @functools.wraps(func)
        async def decorator(*args, **kwargs):
            return await func(*args, **kwargs)

        return decorator
    return wrapper


from .transitions import (
    wait_to_wait_transition,
    wait_to_analyze_transition,
    wait_to_checkpoint_transition,
    wait_to_complete_transition,
    wait_to_error_transition,
    wait_to_execute_transition,
    wait_to_optimize_transition,
    wait_to_setup_transition,
    wait_to_submit_transition,
    wait_to_teardown_transition,
    wait_to_validate_transition
)


import json
import uuid
from typing import List
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet


try:

    import aioredis
    from .redis_config import RedisConfig
    has_connector = True

except Exception:
    aioredis = None
    RedisConfig = None
    has_connector = True


class Redis:

    def __init__(self, config: RedisConfig) -> None:
        self.host = config.host
        self.base = 'rediss' if config.secure else 'redis'
        self.username = config.username
        self.password = config.password
        self.database = config.database
        self.events_channel = config.events_channel
        self.metrics_channel = config.metrics_channel
        self.shared_metrics_channel = f'{self.metrics_channel}_metrics'
        self.errors_channel = f'{self.metrics_channel}_errors'

        self.channel_type = config.channel_type
        self.connection = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to Redis instance at - {self.base}://{self.host} - Database: {self.database}')
        self.connection = await aioredis.from_url(
            f'{self.base}://{self.host}',
            username=self.username,
            password=self.password,
            db=self.database
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to Redis instance at - {self.base}://{self.host} - Database: {self.database}')

    async def submit_events(self, events: List[BaseEvent]):

        if self.channel_type == 'channel':
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Channel - {self.errors_channel}')
            
            for event in events:
                await self.connection.publish(
                    self.events_channel,
                    json.dumps(event.record)
                )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Channel - {self.errors_channel}')

        else:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Redis Set - {self.errors_channel}')
            for event in events:
                await self.connection.sadd(
                    self.events_channel,
                    json.dumps(events.record)
                )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Redis Set - {self.errors_channel}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            if self.channel_type == 'channel':
                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Channel - {self.shared_metrics_channel}')
                await self.connection.publish(
                    self.shared_metrics_channel,
                    json.dumps({
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'group': 'common',
                        **metrics_set.common_stats
                    })
                )

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Channel - {self.shared_metrics_channel}')

            else:
                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Redis Set - {self.shared_metrics_channel}')
                await self.connection.sadd(
                    self.shared_metrics_channel,
                    json.dumps({
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        **metrics_set.common_stats
                    })
                )

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Redis Set - {self.shared_metrics_channel}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                if self.channel_type == 'channel':
                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Channel - {self.metrics_channel} - Group: {group_name}')
                    await self.connection.publish(
                        self.metrics_channel,
                        json.dumps({
                            **group.record,
                            'group': group_name
                        })
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Channel - {self.metrics_channel} - Group: {group_name}')

                else:
                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Redis Set - {self.metrics_channel} - Group: {group_name}')
                    await self.connection.sadd(
                        self.metrics_channel,
                        json.dumps({
                            **group.record,
                            'group': group_name
                        })
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Redis Set - {self.metrics_channel} - Group: {group_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_group_name, group in metrics_set.custom_metrics.items():

                custom_metrics_channel_name = f'{custom_group_name}_metrics'

                if self.channel_type == 'channel':
                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to Channel - {custom_metrics_channel_name} - Group: {custom_group_name}')
                    await self.connection.publish(
                        custom_metrics_channel_name,
                        json.dumps({
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': custom_group_name,
                            **group
                        })
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Channel - {custom_metrics_channel_name} - Group: {custom_group_name}')

                else:
                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to Redis Set - {custom_metrics_channel_name} - Group: {custom_group_name}')
                    await self.connection.sadd(
                        custom_metrics_channel_name,
                        json.dumps({
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': custom_group_name,
                            **group
                        })
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Redis Set - {custom_metrics_channel_name} - Group: {custom_group_name}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):
        
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:

                if self.channel_type == 'channel':
                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Channel - {self.errors_channel}')
                    await self.connection.publish(
                        self.errors_channel,
                        json.dumps({
                            'metrics_name': metrics_set.name,
                            'metrics_stage': metrics_set.stage,
                            'errors_message': error.get('message'),
                            'errors_count': error.get('count')
                        })
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Channel - {self.errors_channel}')

                else:
                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Reids Set - {self.errors_channel}')
                    await self.connection.sadd(
                        self.errors_channel,
                        json.dumps({
                            'metrics_name': metrics_set.name,
                            'metrics_stage': metrics_set.stage,
                            'errors_message': error.get('message'),
                            'errors_count': error.get('count')
                        })
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Reids Set - {self.errors_channel}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connectiion to Redis at - {self.base}://{self.host}')

        await self.connection.close()

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Session Closed - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connectiion to Redis at - {self.base}://{self.host}')